// This file is part of the pdr/pdx project.
// Copyright (C) 2010 Torsten Mueller, Bern, Switzerland
//
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation, either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include "common.h"

using namespace std;
using namespace boost;
using namespace boost::posix_time;
using namespace boost::gregorian;
using namespace boost::program_options;
using namespace boost::filesystem;

#include "xception.h"
#include "config.h"
#include "ascii_with_codepage.h"

#include <Poco/ASCIIEncoding.h>
#include <Poco/Latin1Encoding.h>
#include <Poco/Latin9Encoding.h>
#include <Poco/UTF8Encoding.h>
#include <Poco/UTF16Encoding.h>
#include <Poco/Windows1252Encoding.h>
#include <Poco/TextConverter.h>
#include <Poco/StreamConverter.h>

static string native_encoding(
#ifdef _WIN32
	"ASCII/Codepage 437"
#else
	"UTF-8"
#endif
);

//=== Config ===============================================================
Config::Config ()
	: m_app_name()
	, m_config_file()
	, m_od_visible(256)
	, m_od_cmdline()
	, m_od_cfgfile()
	, m_pod()
	, m_vm()
	, m_now(second_clock::local_time())
{
	Poco::TextEncoding& g = Poco::TextEncoding::global();
	g.add(Poco::SharedPtr<Poco::TextEncoding>(new Poco::ASCIIEncoding()));
	g.add(Poco::SharedPtr<Poco::TextEncoding>(new Poco::Latin1Encoding()));
	g.add(Poco::SharedPtr<Poco::TextEncoding>(new Poco::Latin9Encoding()));
	g.add(Poco::SharedPtr<Poco::TextEncoding>(new Poco::UTF8Encoding()));
	g.add(Poco::SharedPtr<Poco::TextEncoding>(new Poco::UTF16Encoding()));
	g.add(Poco::SharedPtr<Poco::TextEncoding>(new Poco::Windows1252Encoding()));
	g.add(Poco::SharedPtr<Poco::TextEncoding>(new ASCIICodepage437Encoding()));
	g.add(Poco::SharedPtr<Poco::TextEncoding>(new ASCIICodepage850Encoding()));
}

Config::~Config ()
{
}

void Config::Init (int argc, char* argv[]) throw (Xception)
{
	try
	{
		// get the application's name and let derived classes
		// handle it
		m_app_name = path(argv[0]).stem();
		HandleApplicationName();

		// evaluate command line
		command_line_parser cp(argc, argv);
		m_od_cmdline.add(m_od_visible); // a bit tricky
		cp.options(m_od_cmdline);
		cp.positional(m_pod);
		parsed_options po(cp.run());

		ModifyCmdLineOptions(po);

		store(po, m_vm);
	}
	CATCH_RETHROW("error in command line options")

	try
	{
		// evaluate environment (HOME)
		{
			struct X {
				static string func (const string& var)
				{
					return (var == "HOME") ? "home" : "";
				}
			};
			options_description od;
			od.add_options() ("home", "") ;
			store(parse_environment(od, X::func), m_vm);
		}

		// get configuration filename
		m_config_file = GetFilenameOption("config");
		if (m_config_file.empty())
		{
			// .pdrxrc in current directory?
			path p(".");
			p /= ".pdrxrc";
			if (ifstream(p.file_string().c_str()).good())
				m_config_file = p.file_string();
			else
			{
				// .pdrxrc in ~?
				p = GetStringOption("home");
				p /= ".pdrxrc";
				if (ifstream(p.file_string().c_str()).good())
					m_config_file = p.file_string();
				else
				{
					// pdrxrc in ~/.config/pdrx?
					p = GetStringOption("home");
					p /= ".config";
					p /= "pdrx";
					p /= "pdrxrc";
					if (ifstream(p.file_string().c_str()).good())
						m_config_file = p.file_string();
				}
			}
		}

		// evaluate configuration file
		ifstream ifs(m_config_file.c_str(), ios::in);
		if (ifs.good())
		{
			// we still need a vector option for the defines time periods
			// (only in the cfg file)
			m_od_cfgfile.add_options() ("period", value<vector<string> >(), "");
			m_pod.add("period", -1); // unlimited

			parsed_options po(&m_od_cfgfile); // note: this is extending, see below
			string line;
			while (getline(ifs, line))
			{
				static regex rx_cfgfileline("^\\s*([A-Za-z0-9_.-]+)\\s*=\\s*(.*)\\s*$");
				smatch mr;
				if (regex_match(line, mr, rx_cfgfileline, boost::match_not_dot_newline))
				{
					string option_name(mr[1]);

					string option_value(mr[2]);
					trim(option_value);
					if (!option_value.empty() && option_value[0] == '"')
						option_value.erase(0, 1);
					if (!option_value.empty() && option_value[option_value.length() - 1] == '"')
						option_value.erase(option_value.length() - 1, 1);

					bool found = false;
					if (m_od_cfgfile.find_nothrow(option_name, false))
					{
						foreach (option& o, po.options)
						{
							if (o.string_key == option_name)
							{
								o.value.push_back(option_value);
								found = true;
								break;
							}
						}
					}

					if (!found)
					{
						// we extend m_od_cfgfile for every "unknown" option
						if (!m_od_cfgfile.find_nothrow(option_name, false))
							m_od_cfgfile.add_options() (option_name.c_str(), value<string>(), "");
						vector<string> values;
						values.push_back(option_value);
						po.options.push_back(option(option_name, values));
					}
				}
			}

			ModifyCfgFileOptions(po);

			store(po, m_vm);
		}

		notify(m_vm);

		// check if we have a -n parameter
		const string& n = GetStringOption("now");
		if (!n.empty())
		{                          //  CCYY     -MM      -DD     [-hh      :mm      [:ss]]
			static const regex rx("([0-9]+)(-[0-9]+)(-[0-9]+)(-[0-9]+\\:[0-9]+(\\:[0-9]+)?)?");
			smatch mr;
			regex_match(n, mr, rx);

			string t(n);
			if (mr.length(5) > 0)
				;
			else
			{
				if (mr.length(4) > 0)
					t += ":00"; // add seconds
				else
					t += " 00:00:00"; // add complete time
			}
			t += ".000"; // add milliseconds
			t[10] = ' ';

			m_now = time_from_string(t);
		}

		// check if we have a --debug parameter
		if (HasOption("debug"))
			Xception::SetDetailedMode();

		// determine the native encoding
		const string& encoding = GetStringOption("encoding");
		if (!encoding.empty())
		{
			SpecificEncoding(encoding); // provocate exception
			native_encoding = encoding;
		}
	}
	CATCH_RETHROW("error in configuration file")
}

void Config::HandleApplicationName ()
{
}

void Config::ModifyCmdLineOptions (parsed_options& po)
{
}

void Config::ModifyCfgFileOptions (parsed_options& po)
{
}

void Config::Help (ostream& os) const
{
	m_od_visible.print(os);
}

bool Config::HasOption (const string& option) const
{
	try
	{
		return !m_vm[option].empty();
	}
	catch (...)
	{
	}

	return false;
}

bool Config::GetBoolOption (const string& option) const
{
	try
	{
		const variable_value& v = m_vm[option];
		if (!v.empty())
		{
			const any& a = v.value();
			if (a.type() == typeid(string))
			{
				string s(any_cast<string>(a));
				trim(s);
				to_upper(s);
				return s.empty() || s == "YES" || s == "TRUE" || s == "1";
			}
			else
			{
				if (a.type() == typeid(bool))
					return any_cast<bool>(a);
			}
		}
	}
	catch (...)
	{
	}

	return false;
}

string Config::GetStringOption (const string& option) const
{
	try
	{
		const variable_value& v = m_vm[option];
		if (!v.empty())
			return any_cast<string>(v.value());
	}
	catch (...)
	{
	}

	return string();
}

double Config::GetDoubleOption (const string& option) const
{
	try
	{
		const variable_value& v = m_vm[option];
		if (!v.empty())
			return lexical_cast<double>(any_cast<string>(v.value()));
	}
	catch (...)
	{
	}

	return 0.0;
}

vector<string> Config::GetVectorOption (const string& option) const
{
	try
	{
		const variable_value& v = m_vm[option];
		if (!v.empty())
			return any_cast<vector<string> >(v.value());
	}
	catch (...)
	{
	}

	return vector<string>();
}

string Config::GetFilenameOption (const string& option) const
{
	try
	{
		const variable_value& v = m_vm[option];
		if (!v.empty())
		{
			string filename(any_cast<string>(v.value()));
			if (filename.find("~/") == 0)
			{
				filename.erase(0, 2);
				path p(any_cast<string>(m_vm["home"].value()));
				p /= filename;
				filename = p.file_string();
			}
			return filename;
		}
	}
	catch (...)
	{
	}

	return string();
}

const string& Config::GetApplicationName () const
{
	return m_app_name;
}

ptime Config::GetNow () const
{
	// in fast mode we need constant value for "now" for getting usable
	// keys in the selection cache, if we are not fast, we work with
	// real timestamps

	// m_now is thought to be an option in future

	return (GetBoolOption("fast") || !GetStringOption("now").empty()) ? m_now : second_clock::local_time();
}

const string& Config::GetConfigFile () const
{
	return m_config_file;
}

//=== encoding stuff =======================================================
void GetEncodingNames (EncodingNames& names)
{
	static const char* n[] = {
		"ASCII",
		"ISO-8859-1",	"Latin1",	"Latin-1",
		"ISO-8859-15",	"Latin9",	"Latin-9",
		"UTF-8",	"UTF8",
		"UTF-16",	"UTF16",
		"Windows-1252",	"CP1252",
		"ASCII/437",	"ASCII/CP437",	"ASCII/Codepage 437",
		"ASCII/850",	"ASCII/CP850",	"ASCII/Codepage 850"
	};

	names.clear();
	for (size_t i = 0; i < sizeof(n) / sizeof(const char*); i++)
		names.insert(n[i]);
}

Poco::TextEncoding& InternalEncoding ()
{
	return Poco::TextEncoding::global();
}

Poco::TextEncoding& NativeEncoding ()
{
	return *Poco::TextEncoding::global().find(native_encoding);
}

Poco::TextEncoding& SpecificEncoding (const string& encoding) throw (Xception)
{
	Poco::TextEncoding::Ptr p = Poco::TextEncoding::global().find(encoding);
	if (p)
		return *p;
	else
		THROW(format("unknown encoding: ") % encoding);
}

string Decode (const string& text, Poco::TextEncoding& from)
{
	Poco::TextEncoding& g = Poco::TextEncoding::global();
	if (&from == &g)
		return text;
	else
	{
		string s;
		Poco::TextConverter(from, g).convert(text, s);
		return s;
	}
}

string Encode (const string& text, Poco::TextEncoding& to)
{
	Poco::TextEncoding& g = Poco::TextEncoding::global();
	if (&to == &g)
		return text;
	else
	{
		string s;
		Poco::TextConverter(g, to).convert(text, s);
		return s;
	}
}
