// This file is part of the pdr/pdx project.
// Copyright (C) 2010 Torsten Mueller, Bern, Switzerland
//
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation, either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include "../libpdrx/common.h"

using namespace std;
using namespace boost;
using namespace boost::posix_time;
using namespace boost::gregorian;
using namespace boost::program_options;

#include "../libpdrx/xception.h"
#include "../libpdrx/conversions.h"
#include "../libpdrx/config.h"
#include "db.h"
#include "in_impl.h"

#include <Poco/Net/NetException.h>
#include <Poco/Net/POP3ClientSession.h>
#include <Poco/Net/MailMessage.h>
#include <Poco/Net/QuotedPrintableDecoder.h>

using namespace Poco;
using namespace Poco::Net;

//=== Pop3MailClient =======================================================
Pop3MailClient::Pop3MailClient (const string& option_key)
	: InputImpl(option_key)
{
}

	static void decode_quoted_printable (string& s)
	{
		stringstream ss(s);
		QuotedPrintableDecoder decoder(ss);
		string decoded;
		decoder >> decoded;
		s = decoded;
	}

	static void convert_charset (string& s, const string& from_charset)
	{
		s = Decode(s, SpecificEncoding(from_charset));
	}

void Pop3MailClient::Do (const Config& config, Database& database) const throw (Xception)
{
	try
	{
		bool verbose = config.GetBoolOption("verbose");
		if (verbose)
			encoded::cout << "looking for mail (POP3)" << endl;

		// get configuration data
		const string& server = config.GetStringOption(m_option_key + ".server");
		if (server.empty())
			THROW(format("missing specification in configuration file: %s.server)") % m_option_key);
		const string& account = config.GetStringOption(m_option_key + ".account");
		if (account.empty())
			THROW(format("missing specification in configuration file: %s.account)") % m_option_key);
		const string& password = config.GetStringOption(m_option_key + ".password");
		const string& subject = config.GetStringOption(m_option_key + ".subject");
		bool keep = config.GetBoolOption(m_option_key + ".keep");

		// open connection and log in
		POP3ClientSession session(server);
		session.login(account, password);

		// process available mails
		POP3ClientSession::MessageInfoVec messages;
		session.listMessages(messages);

		Database::Collections collections;
		database.GetCollections(collections);

		bool data = false;
		bool rejects = false;
		POP3ClientSession::MessageInfoVec::const_iterator I = messages.begin();
		while (I != messages.end())
		{
			int id = (*I++).id;

			MailMessage message;
			session.retrieveMessage(id, message);

			// check subject
			string s(message["Subject"]);
			{
				trim(s);
				s = InputImpl::ConvertMailSubject(s, decode_quoted_printable, convert_charset);
			}

			if (s != subject)
				continue;

			data = true;

			// extract timestamp
			ptime timestamp(not_a_date_time);
			{
				string d(message["Date"]);
				trim(d);
				timestamp = lexical_cast_mime_date(d);
			}

			// get message body and decode it correctly (this means make it UTF-8)
			const string& charset = ExtractCharsetFromContentType(message.getContentType());
			stringstream ss;
			if (charset.empty())
				ss << message.getContent(); // should be 7bit ASCII
			else
				ss << Decode(message.getContent(), SpecificEncoding(charset));

			// parse line by line, every line an expression
			string line;
			while (getline(ss, line))
			{
				trim(line);
				if (line.empty())
					continue;

				try
				{
					Database::CollectionsItems items;
					Parse(line, timestamp, verbose, collections, items);
					database.AddCollectionsItems(items);
				}
				catch (const Xception& x)
				{
					encoded::cerr << x.Message(Xception::brief) << endl;
					database.AddRejected(timestamp, line);
					rejects = true;
				}
			}

			// delete if neccessary
			if (!keep)
				session.deleteMessage(id);
		}
		if (verbose && !data)
			encoded::cout << "    no data on server" << endl;

		// close connection
		session.close();

		if (rejects)
			encoded::cerr << "!!! at least one expression has been rejected, try -r to list rejections !!!" << endl;
	}
	CATCH_RETHROW("could not retrieve mail data from pop3 server")
}
